/***************************************************************************
 *   Copyright (C) 2004 by Francisco J. Ros                                *
 *   fjrm@dif.um.es                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
*	Modified by: Miguel Luís			2009								*
* 					nml17338@fct.unl.pt										*
* 				 Rodolfo Oliveira		2009								*	
* 					rado@uninova.pt											*
* 																			*
* 	This new version of OLSR protocol was modified to improve the OLSR		*
* 	performance in VANETs. The new version is called OLSR-FCT.				*
* 																			*
*  Faculty of Sciences and Technology (FCT) - New University of Lisbon (UNL)*
 ***************************************************************************/

/*!
 * \mainpage
 *
 * \author Francisco J. Ros
 * \date 11 March, 2005
 *
 * \section feat Features
 *
 * <p>This software implements base specification of <a href="http://www.ietf.org/rfc/rfc3626.txt">
 *    %OLSR protocol</a> for <a href="http://www.isi.edu/nsnam/ns/">NS2 network simulator</a>.
 *    It has been developed at the <a href="http://www.um.es/">University of Murcia</a>
 *    (Spain) by <a href="http://ants.dif.um.es/~fjrm/">Francisco J. Ros</a>, and it is released
 *    under the <a href="http://www.gnu.org/copyleft/gpl.html">GNU General Public License</a>.</p>
 *
 * <p>There are available patches for the latest releases of <strong>NS2</strong>, namely for
 *    versions 2.27 and 2.28. Tests have been performed on a Debian GNU/Linux system, but the
 *    code should work on all platforms supported by <strong>NS2</strong>.</p>
 *
 * <p>Here is a summary of software's main features:</p>
 *
 * <ul>
 *  <li>Compliant with core-OLSR (but the use of multiple interfaces which is currently
 *      not supported by <strong>NS2</strong>) as it is documented in RFC 3626.</li>
 *
 *  <li>Support for MAC layer feedback as described in RFC 3626.</li>
 *
 *  <li>Source code follows <strong>NS2</strong> coding style.</li>
 *
 *  <li>Highly configurable from TCL scripts, i.e., without the need of
 *      recompiling the whole simulator. You can:</li>
 *
 *  <ul>
 *    <li>Activate/deactivate debug mode.</li>
 *
 *    <li>Change the interval at which every message type is sent (on a
 *        per-node basis).</li>
 *
 *    <li>Change a node's willingness for forwarding data packets on
 *        behalf of other nodes (on a per-node basis).</li>
 *
 *    <li>Print whatever data structure managed by a node at a certain
 *        time (on a per-node basis).</li>
 *  </ul>
 *
 * </ul>
 *
 * \section inst Installing
 *
 * <p>In this section we are supposing that you have installed a version of <strong>NS</strong>
 *    (2.27/2.28) which is properly running. Let's call <em>$NS_ROOT</em> the upper directory
 *    where <strong>NS</strong> is installed. For instance, if you download and install
 *    <em>ns-allinone-2.28.tar.gz</em> in your home directory, then <em>$NS_ROOT</em> is
 *    <em>~/ns-allinone-2.28</em>. Besides let's call <em>$VERSION</em> to the version of
 *    your installed <strong>NS</strong> software.</p>
 *
 * <p>There are two options for installing <strong>UM-OLSR</strong>, either applying a patch
 *    or copying each source file to the appropiate place:</p>
 *
 * <ul>
 *  <li>Applying patches</li>
 *
 *  <ol>
 *   <li>Download <strong>UM-OLSR</strong> patch for <strong>NS</strong>
 *       <a href="http://ants.dif.um.es/masimum/um-olsr/um-olsr_ns-2.27_v0.8.7.patch">2.27</a> or
 *       <a href="http://ants.dif.um.es/masimum/um-olsr/um-olsr_ns-2.28_v0.8.7.patch">2.28</a>
 *       version. Let's call <em>$PATCH</em> to the downloaded file.</li>
 *
 *   <li><tt># cp $PATCH $NS_ROOT && cd $NS_ROOT</tt></li>
 *
 *   <li><tt># patch -p0 \< $PATCH</tt></li>
 *
 *   <li><tt># cd ns-$VERSION</tt></li>
 *
 *   <li><tt># make</tt></li>
 *  </ol>
 *
 *  <li>Copying files</li>
 *
 *  <ol>
 *   <li>Download <strong>UM-OLSR</strong> code for <strong>NS</strong>
 *       <a href="http://ants.dif.um.es/masimum/um-olsr/um-olsr_ns-2.27_v0.8.7.tar.gz">2.27</a> or
 *       <a href="http://ants.dif.um.es/masimum/um-olsr/um-olsr_ns-2.28_v0.8.7.tar.gz">2.28</a>
 *       version. Let's call <em>$FILE</em> to the downloaded file.</li>
 *
 *   <li><tt># tar zxvf $FILE</tt></li>
 *
 *   <li>Under the recently created directory are several files which must be copied
 *       to the corresponding places in <em>$NS_ROOT</em>/ns-<em>$VERSION</em>.</li>
 *
 *   <li><tt># cd $NS_ROOT/ns-$VERSION</tt></li>
 *
 *   <li><tt># make</tt></li>
 *  </ol>
 *
 * \section howto How to Use It
 *
 * <p><strong>UM-OLSR</strong> allows <strong>NS2</strong> users to use %OLSR protocol like
 *    any other MANET routing protocol available in the official release.</p>
 *
 * <p><a href="http://ants.dif.um.es/masimum/um-olsr/olsr_example.tcl">This example</a>
 * explores main features of <strong>UM-OLSR</strong> and shows how to perform simulations
 * with this routing protocol. If you can't understand the content of the file, then you
 * should start reading <a href="http://www.isi.edu/nsnam/ns/ns-documentation.html">NS
 * Manual</a> or any of the available tutorials for beginners (see
 * <a href="http://www.isi.edu/nsnam/ns/">NS web page</a>). The simulation consists of a
 * simple TCL script which defines five mobile nodes (with no motion) and an UDP connection
 * between two of them.</p>
 *
 * \section traces Understanding Trace Format
 *
 * <strong>UM-OLSR</strong> supports all trace formats defined in <strong>NS2</strong>,
 * i.e, old, new and tagged trace formats. In all cases it is shown the same information
 * when an OLSR packet is sent, received, dropped or forwarded:
 *
 * <ul>
 *  <li>Number of messages this packet contains.</li>
 *  <li>Packet sequence number.</li>
 *  <li>For each message is also shown:</li>
 *   <ul>
 *    <li>Message type ("HELLO", "TC" or "UNKNOWN").</li>
 *    <li>Originator address.</li>
 *    <li>Hop count.</li>
 *    <li>Message sequence number.</li>
 *   </ul>
 * </ul>
 *
 * If you need a more detailed description of trace format, see <em>CMUTrace::format_olsr()</em>
 * function in <em>$NS_ROOT/ns-$VERSION/trace/cmu-trace.cc</em> or the result of the above mentioned
 * <a href="http://ants.dif.um.es/masimum/um-olsr/olsr_example.tcl">example</a>.
 */

///
/// \file	OLSR_FCT.h
/// \brief	Header file for OLSR agent and related classes.
///
/// Here are defined all timers used by OLSR, including those for managing internal
/// state and those for sending messages. Class OLSR is also defined, therefore this
/// file has signatures for the most important methods. Lots of constants are also
/// defined.
///

#ifndef __OLSR_FCT_h__
#define __OLSR_FCT_h__


#include <olsr_FCT/OLSR_FCT_pkt.h>
#include <olsr_FCT/OLSR_FCT_state.h>
#include <olsr_FCT/OLSR_FCT_rtable.h>
#include <olsr_FCT/OLSR_FCT_repositories.h>
#include <trace.h>
#include <classifier-port.h>
#include <agent.h>
#include <packet.h>
#include <timer-handler.h>
#include <random.h>
#include <vector>

#include <olsr_FCT/cbc_table.h>

/********** Useful macros **********/

/// Returns maximum of two numbers.
#ifndef MAX
#define	MAX(a, b) (((a) > (b)) ? (a) : (b))
#endif

/// Returns minimum of two numbers.
#ifndef MIN
#define	MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif

/// Defines NULL as zero, used for pointers.
#ifndef NULL
#define NULL 0
#endif

/// Gets current time from the scheduler.
#define CURRENT_TIME	Scheduler::instance().clock()

///
/// \brief Gets the delay between a given time and the current time.
///
/// If given time is previous to the current one, then this macro returns
/// a number close to 0. This is used for scheduling events at a certain moment.
///
#define DELAY(time) (((time) < (CURRENT_TIME)) ? (0.000001) : \
	(time - CURRENT_TIME + 0.000001))

/// Scaling factor used in RFC 3626.
#define OLSR_C		0.0625


/********** Intervals **********/

/// HELLO messages emission interval.
#define OLSR_HELLO_INTERVAL	2

/// TC messages emission interval.
#define OLSR_TC_INTERVAL	5

/// MID messages emission interval.
#define OLSR_MID_INTERVAL	OLSR_TC_INTERVAL

///
/// \brief Period at which a node must cite every link and every neighbor.
///
/// We only use this value in order to define OLSR_NEIGHB_HOLD_TIME.
///
#define OLSR_REFRESH_INTERVAL	2


/********** Holding times **********/

/// Neighbor holding time.
#define OLSR_NEIGHB_HOLD_TIME	3*OLSR_REFRESH_INTERVAL
/// Top holding time.
#define OLSR_TOP_HOLD_TIME	3*OLSR_TC_INTERVAL
/// Dup holding time.
#define OLSR_DUP_HOLD_TIME	30
/// MID holding time.
#define OLSR_MID_HOLD_TIME	3*OLSR_MID_INTERVAL


/********** Link types **********/

/// Unspecified link type.
#define OLSR_UNSPEC_LINK	0
/// Asymmetric link type.
#define OLSR_ASYM_LINK		1
/// Symmetric link type.
#define OLSR_SYM_LINK		2
/// Lost link type.
#define OLSR_LOST_LINK		3

/********** Neighbor types **********/

/// Not neighbor type.
#define OLSR_NOT_NEIGH		0
/// Symmetric neighbor type.
#define OLSR_SYM_NEIGH		1
/// Asymmetric neighbor type.
#define OLSR_MPR_NEIGH		2


/********** Willingness **********/

/// Willingness for forwarding packets from other nodes: never.
#define OLSR_WILL_NEVER		0
/// Willingness for forwarding packets from other nodes: low.
#define OLSR_WILL_LOW		1
/// Willingness for forwarding packets from other nodes: medium.
#define OLSR_WILL_DEFAULT	3
/// Willingness for forwarding packets from other nodes: high.
#define OLSR_WILL_HIGH		6
/// Willingness for forwarding packets from other nodes: always.
#define OLSR_WILL_ALWAYS	7


/********** Miscellaneous constants **********/

/// Maximum allowed jitter.
#define OLSR_MAXJITTER		OLSR_HELLO_INTERVAL/8
/// Maximum allowed sequence number.
#define OLSR_MAX_SEQ_NUM	65535
/// Used to set status of an OLSR_nb_tuple as "not symmetric".
#define OLSR_STATUS_NOT_SYM	0
/// Used to set status of an OLSR_nb_tuple as "symmetric".
#define OLSR_STATUS_SYM		1
/// Random number between [0-OLSR_MAXJITTER] used to jitter OLSR packet transmission.
#define JITTER			(Random::uniform()*OLSR_MAXJITTER)


class OLSR_FCT;			// forward declaration


/********** Timers **********/


/// Timer for sending an enqued message.
class OLSR_FCT_MsgTimer : public TimerHandler {
public:
	OLSR_FCT_MsgTimer(OLSR_FCT* agent) : TimerHandler() { agent_	= agent; }
protected:
	OLSR_FCT*	agent_;			///< OLSR agent which created the timer.
	virtual void expire(Event* e);
};

/// Timer for sending HELLO messages.
class OLSR_FCT_HelloTimer : public TimerHandler {
public:
	OLSR_FCT_HelloTimer(OLSR_FCT* agent) : TimerHandler() { agent_ = agent; }
protected:
	OLSR_FCT*	agent_;			///< OLSR agent which created the timer.
	virtual void expire(Event* e);
};


/// Timer for sending TC messages.
class OLSR_FCT_TcTimer : public TimerHandler {
public:
	OLSR_FCT_TcTimer(OLSR_FCT* agent) : TimerHandler() { agent_ = agent; }
protected:
	OLSR_FCT*	agent_;			///< OLSR agent which created the timer.
	virtual void expire(Event* e);
};


/// Timer for sending MID messages.
class OLSR_FCT_MidTimer : public TimerHandler {
public:
	OLSR_FCT_MidTimer(OLSR_FCT* agent) : TimerHandler() { agent_ = agent; }
protected:
	OLSR_FCT*	agent_;			///< OLSR agent which created the timer.
	virtual void expire(Event* e);
};


/// Timer for removing duplicate tuples: OLSR_dup_tuple.
class OLSR_FCT_DupTupleTimer : public TimerHandler {
public:
	OLSR_FCT_DupTupleTimer(OLSR_FCT* agent, OLSR_dup_tuple* tuple) : TimerHandler() {
		agent_ = agent;
		tuple_ = tuple;
	}
protected:
	OLSR_FCT*		agent_;	///< OLSR agent which created the timer.
	OLSR_dup_tuple*	tuple_;	///< OLSR_dup_tuple which must be removed.
	
	virtual void expire(Event* e);
};


/// Timer for removing link tuples: OLSR_link_tuple.
class OLSR_FCT_LinkTupleTimer : public TimerHandler {
	///
	/// \brief A flag which tells if the timer has expired (at least) once or not.
	///
	/// When a link tuple has been just created, its sym_time is expired but this
	/// does not mean a neighbor loss. Thus, we use this flag in order to be able
	/// to distinguish this situation.
	///
	bool			first_time_;
public:
	OLSR_FCT_LinkTupleTimer(OLSR_FCT* agent, OLSR_link_tuple* tuple) : TimerHandler() {
		agent_		= agent;
		tuple_		= tuple;
		first_time_	= true;
	}
protected:
	OLSR_FCT*			agent_;	///< OLSR agent which created the timer.
	OLSR_link_tuple*	tuple_;	///< OLSR_link_tuple which must be removed.
	
	virtual void expire(Event* e);
};


/// Timer for removing nb2hop tuples: OLSR_nb2hop_tuple.
class OLSR_FCT_Nb2hopTupleTimer : public TimerHandler {
public:
	OLSR_FCT_Nb2hopTupleTimer(OLSR_FCT* agent, OLSR_nb2hop_tuple* tuple) : TimerHandler() {
		agent_ = agent;
		tuple_ = tuple;
	}
protected:
	OLSR_FCT*			agent_;	///< OLSR agent which created the timer.
	OLSR_nb2hop_tuple*	tuple_;	///< OLSR_nb2hop_tuple which must be removed.
	
	virtual void expire(Event* e);
};


/// Timer for removing MPR selector tuples: OLSR_mprsel_tuple.
class OLSR_FCT_MprSelTupleTimer : public TimerHandler {
public:
	OLSR_FCT_MprSelTupleTimer(OLSR_FCT* agent, OLSR_mprsel_tuple* tuple) : TimerHandler() {
		agent_ = agent;
		tuple_ = tuple;
	}
protected:
	OLSR_FCT*			agent_;	///< OLSR agent which created the timer.
	OLSR_mprsel_tuple*	tuple_;	///< OLSR_mprsel_tuple which must be removed.
	
	virtual void expire(Event* e);
};


/// Timer for removing topology tuples: OLSR_topology_tuple.
class OLSR_FCT_TopologyTupleTimer : public TimerHandler {
public:
	OLSR_FCT_TopologyTupleTimer(OLSR_FCT* agent, OLSR_topology_tuple* tuple) : TimerHandler() {
		agent_ = agent;
		tuple_ = tuple;
	}
protected:
	OLSR_FCT*			agent_;	///< OLSR agent which created the timer.
	OLSR_topology_tuple*	tuple_;	///< OLSR_topology_tuple which must be removed.
	
	virtual void expire(Event* e);
};


/// Timer for removing interface association tuples: OLSR_iface_assoc_tuple.
class OLSR_FCT_IfaceAssocTupleTimer : public TimerHandler {
public:
	OLSR_FCT_IfaceAssocTupleTimer(OLSR_FCT* agent, OLSR_iface_assoc_tuple* tuple) : TimerHandler() {
		agent_ = agent;
		tuple_ = tuple;
	}
protected:
	OLSR_FCT*			agent_;	///< OLSR agent which created the timer.
	OLSR_iface_assoc_tuple*	tuple_;	///< OLSR_iface_assoc_tuple which must be removed.
	
	virtual void expire(Event* e);
};


/********** OLSR Agent **********/


///
/// \brief Routing agent which implements %OLSR protocol following RFC 3626.
///
/// Interacts with TCL interface through command() method. It implements all
/// functionalities related to sending and receiving packets and managing
/// internal state.
///
class OLSR_FCT : public Agent {
	// Makes some friends.
	friend class OLSR_FCT_HelloTimer;
	friend class OLSR_FCT_TcTimer;
	friend class OLSR_FCT_MidTimer;
	friend class OLSR_FCT_DupTupleTimer;
	friend class OLSR_FCT_LinkTupleTimer;
	friend class OLSR_FCT_Nb2hopTupleTimer;
	friend class OLSR_FCT_MprSelTupleTimer;
	friend class OLSR_FCT_TopologyTupleTimer;
	friend class OLSR_FCT_IfaceAssocTupleTimer;
	friend class OLSR_FCT_MsgTimer;
	
	// Modified ML
	Beacons_tab 					*tab_bc;
	std::map<nsaddr_t, u_int16_t> 	node_assoc;
	bool 							online;
	
	std::vector<int> 				is_bgl_time;
	float							coverage;
	// **
	
	/// Address of the routing agent.
	nsaddr_t	ra_addr_;
	
	/// Packets sequence number counter.
	u_int16_t	pkt_seq_;
	/// Messages sequence number counter.
	u_int16_t	msg_seq_;
	/// Advertised Neighbor Set sequence number.
	u_int16_t	ansn_;

	/// HELLO messages' emission interval.
	double		hello_ival_;
	/// TC messages' emission interval.
	double		tc_ival_;
	/// MID messages' emission interval.
	int		mid_ival_;
	/// Willingness for forwarding packets on behalf of other nodes.
	int		willingness_;
	/// Determines if layer 2 notifications are enabled or not.
	int		use_mac_;
	
	/// Routing table.
	OLSR_FCT_rtable		rtable_;
	/// Internal state with all needed data structs.
	OLSR_FCT_state		state_;
	/// A list of pending messages which are buffered awaiting for being sent.
	std::vector<OLSR_FCT_msg>	msgs_;

	///
	/// \brief A list of timers which have expired but their memory haven't been released yet.
	///
	/// This is a very ugly hack. Since we are dynamically creating
	/// several timers we need to release their memory at a certain
	/// time. Here we will store their references in order to later
	/// on release them.
	///
	std::vector<TimerHandler*>	expired_timers_;
	
	void				release_expired_timers();	// See just above.
	
protected:
	PortClassifier*	dmux_;		///< For passing packets up to agents.
	Trace*		logtarget_;	///< For logging.
	
	OLSR_FCT_HelloTimer	hello_timer_;	///< Timer for sending HELLO messages.
	OLSR_FCT_TcTimer	tc_timer_;	///< Timer for sending TC messages.
	OLSR_FCT_MidTimer	mid_timer_;	///< Timer for sending MID messages.

	
	/// Increments packet sequence number and returns the new value.
	inline u_int16_t	pkt_seq() {
		pkt_seq_ = (pkt_seq_ + 1)%(OLSR_MAX_SEQ_NUM + 1);
		return pkt_seq_;
	}
	/// Increments message sequence number and returns the new value.
	inline u_int16_t	msg_seq() {
		msg_seq_ = (msg_seq_ + 1)%(OLSR_MAX_SEQ_NUM + 1);
		return msg_seq_;
	}
	
	inline nsaddr_t&	ra_addr()	{ return ra_addr_; }
	
	inline double&		hello_ival()	{ return hello_ival_; }
	inline double&		tc_ival()	{ return tc_ival_; }
	inline int&		mid_ival()	{ return mid_ival_; }
	inline int&		willingness()	{ return willingness_; }
	inline int&		use_mac()	{ return use_mac_; }
	
	inline linkset_t&	linkset()	{ return state_.linkset(); }
	inline mprset_t&	mprset()	{ return state_.mprset(); }
	inline mprselset_t&	mprselset()	{ return state_.mprselset(); }
	inline nbset_t&		nbset()		{ return state_.nbset(); }
	inline nb2hopset_t&	nb2hopset()	{ return state_.nb2hopset(); }
	inline topologyset_t&	topologyset()	{ return state_.topologyset(); }
	inline dupset_t&	dupset()	{ return state_.dupset(); }
	inline ifaceassocset_t&	ifaceassocset()	{ return state_.ifaceassocset(); }
	
	void		recv_olsr(Packet*);
	
	void		mpr_computation();
	void		rtable_computation();
	
	void		process_hello(OLSR_FCT_msg&, nsaddr_t, nsaddr_t);
	void		process_tc(OLSR_FCT_msg&, nsaddr_t);
	void		process_mid(OLSR_FCT_msg&, nsaddr_t);
	
	void		forward_default(Packet*, OLSR_FCT_msg&, OLSR_dup_tuple*, nsaddr_t);
	void		forward_data(Packet*);
	
	void		enque_msg(OLSR_FCT_msg&, double);
	void		send_hello();
	void		send_tc();
	void		send_mid();
	void		send_pkt();
	
	void		link_sensing(OLSR_FCT_msg&, nsaddr_t, nsaddr_t);
	void		populate_nbset(OLSR_FCT_msg&);
	void		populate_nb2hopset(OLSR_FCT_msg&);
	void		populate_mprselset(OLSR_FCT_msg&);

	void		set_hello_timer();
	void		set_tc_timer();
	void		set_mid_timer();

	void		nb_loss(OLSR_link_tuple*);
	void		add_dup_tuple(OLSR_dup_tuple*);
	void		rm_dup_tuple(OLSR_dup_tuple*);
	void		add_link_tuple(OLSR_link_tuple*, u_int8_t);
	void		rm_link_tuple(OLSR_link_tuple*);
	void		updated_link_tuple(OLSR_link_tuple*);
	void		add_nb_tuple(OLSR_nb_tuple*);
	void		rm_nb_tuple(OLSR_nb_tuple*);
	void		add_nb2hop_tuple(OLSR_nb2hop_tuple*);
	void		rm_nb2hop_tuple(OLSR_nb2hop_tuple*);
	void		add_mprsel_tuple(OLSR_mprsel_tuple*);
	void		rm_mprsel_tuple(OLSR_mprsel_tuple*);
	void		add_topology_tuple(OLSR_topology_tuple*);
	void		rm_topology_tuple(OLSR_topology_tuple*);
	void		add_ifaceassoc_tuple(OLSR_iface_assoc_tuple*);
	void		rm_ifaceassoc_tuple(OLSR_iface_assoc_tuple*);
	
	nsaddr_t	get_main_addr(nsaddr_t);
	int		degree(OLSR_nb_tuple*); 

	static bool	seq_num_bigger_than(u_int16_t, u_int16_t);
	
	// Modified ML 
	u_int8_t 		check_stability();
	u_int8_t		conv_assoc_will(nsaddr_t, u_int8_t, bool);
	double			count_size_N2(nb2hopset_t);
	// **

public:
	OLSR_FCT(nsaddr_t);
	int	command(int, const char*const*);
	void	recv(Packet*, Handler*);
	void	mac_failed(Packet*);
	
	static double		emf_to_seconds(u_int8_t);
	static u_int8_t		seconds_to_emf(double);
	static int		node_id(nsaddr_t);
};

#endif
