/*////////////////////////////////////////////////////////////////////////
//																		//
//  TITLE: cbc_table.cc - Node's beacon table class implementatiofile	//
//																		//
//																		//
//  FUNCTION: Implements all classes responsible for manage				//
//	      the received becons on each node								//
//																		//
//																		//
//  AUTHOR: Rodolfo Oliveira - June 2004								//
//	    rado@uninova.pt													//
//	Modified by: Miguel Luís - September 2009							//
//		nml17338@fct.unl.pt												//
//																		//
//  New University of Lisbon / FCT - All rights reserved				//
//																		//
////////////////////////////////////////////////////////////////////////*/
#include "cbc_table.h"

//#define	DEBUG_CLUSTERING //prints all beacon table for each update

int Beacons_tab::link_counter; 			// Total lynks ocurred 
vector_lst Beacons_tab::histogram; 		// Histogram with number of cleared links


//------------------------------------------------------------------------
//- class Addr_list_node
Addr_list_node::Addr_list_node() 
{
	addr = -1; 
	next = NULL;
}


//------------------------------------------------------------------------
//- class Addr_list
Addr_list:: Addr_list()
{
	// init
	m_head = NULL;
	m_tail = NULL;
	m_elems = 0;
}

Addr_list::~Addr_list()
{
	this->Delete_all();
}

int Addr_list::Delete_all()
{
	Addr_list_node *ptr, *next;

	ptr = m_head;
	if (m_elems != 0)
		for (int i =0; i<m_elems; i++) {
			next = ptr->next;
			delete ptr;
			ptr = next;
		}
	return 0;
}

void Addr_list::Add_elem(nsaddr_t *address)
{
	Addr_list_node* new_node = (Addr_list_node*) new Addr_list_node();
	new_node->addr = *address;
	m_elems++;
	if (m_elems == 1) {
		m_head = new_node;
		m_tail = new_node;
		return;
	}
	m_tail->next = new_node;
	m_tail = new_node;
}

int Addr_list::Get_num_elems()
{
	return m_elems;
}

int Addr_list::Get_elem_at(int pos, nsaddr_t *result)
{
	Addr_list_node *ptr;
	if (pos > m_elems-1)
		return -1;
	ptr = m_head;
	for (int i =0; i< pos; i++) // position starts at index 0
		ptr = ptr->next;
	*result = ptr->addr;
	return 0;
}

void Addr_list::Charge_list(char* ptr, u_int8_t bc_gl_lis_elems)
{
	nsaddr_t address;
	u_int8_t  num_elems = bc_gl_lis_elems;
	this->Delete_all();
	for(int i=0; i<num_elems;i++) {
		address = (nsaddr_t) *(ptr+i*sizeof(address));
        this->Add_elem(&address);
	}
}


//------------------------------------------------------------------------
//- class Timer_delete
void Timer_delete::expire(Event *e) {
	a_->tab->Delete_beacon(&(a_->bc_elem_node));
};

//------------------------------------------------------------------------
//- class Timer_bPeriod
void Timer_bPeriod::expire(Event *e) {
   a_->bc_elem_associativity++;
   a_->tab->Refresh_group_leader();
   double b_period=(double)BEACON_AT_INTERVAL + (double)MAJORANTE_JITTER /(double)2.0;
   this->resched(b_period);
};



//------------------------------------------------------------------------
//- class Beacons_tab_node
Beacons_tab_node:: Beacons_tab_node() : timer(this), count_beaconsTimer(this)
{
	bc_elem_node = -1;
	bc_elem_seq_num = 0;
	bc_elem_associativity = 0;
	bc_elem_bl_associativity = 0;
	bc_elem_gl_addr = -1;
	bc_elem_gl_list = (Addr_list*) new Addr_list();
	next = NULL;
}

Beacons_tab_node::~Beacons_tab_node()
{
   // stops beacon counter
   if (this->count_beaconsTimer.status()==1)    
	   this->count_beaconsTimer.cancel();
   delete bc_elem_gl_list;
   // delete all beacons contained in this node
}


//------------------------------------------------------------------------
// - class Beacons_tab
Beacons_tab::Beacons_tab()
{
	m_head = NULL;
	m_tail = NULL;
	m_elems = 0;

	bc_associativity = 0;
	bc_gl_addr = -1;
	bc_gl_lis_elems = 0;	//num of elements contained in bc_gl_list - list of group leaders in the neigborhood
	bc_gl_list = NULL;

	m_is_stable  = false;
	m_is_bgl     = false;

	// link duration statistics
	link_counter = 0;
}

Beacons_tab::~Beacons_tab()
{
	 Beacons_tab_node *ptr, *next;

	 ptr = m_head; 
	 if (m_elems != 0)
		 while(m_elems>0) {
			 next = ptr->next;
			 Delete_beacon(&ptr->bc_elem_node);
			 ptr = next;
		 }
   
	 //print_link_duration_stats();
}

void Beacons_tab::Add_beacon(u_int32_t bc_seq_num, u_int32_t bc_associativity, nsaddr_t bc_gl_addr, nsaddr_t *received_from)
{
	Beacons_tab_node *ptr;
	ptr = m_head; 
	int res = -1;
	// search if address and seq_num received is contained in the table
	for (int i=0; i<m_elems;i++) {
		if (ptr->bc_elem_node == *received_from) {
			if (ptr->bc_elem_seq_num == bc_seq_num) {
				res = 1;
				break;
			}
			else
			{
				res = 0;
				break;
			}
		}
		ptr = ptr->next;
	}

	// insert new node because this address is not inserted in the table
	if (res == -1)
	{
	    Beacons_tab_node* new_node = (Beacons_tab_node*) new Beacons_tab_node();
	    new_node->tab = this; 		// pointer to delete the entry by a timeout....
	    new_node->bc_elem_node = *received_from;
	    new_node->bc_elem_seq_num = bc_seq_num;
	    new_node->bc_elem_associativity++;
	    new_node->bc_elem_bl_associativity = bc_associativity;
	    new_node->bc_elem_gl_addr = bc_gl_addr;
	    new_node->next = NULL;
	    m_elems++;  
  
	    // refresh pointers
	    if (m_elems == 1) {
	    	m_head = new_node;
	    	m_tail = new_node;
	    }		
	    else {
	    	m_tail->next = new_node;
	    	m_tail = new_node;
	    }

	    // a new timer is launched to count each beacon period in order 
	    // to increment ptr->bc_elem_associativity
	    double b_period=(double)BEACON_AT_INTERVAL + (double)MAJORANTE_JITTER /(double)2.0;
	    if (m_tail->count_beaconsTimer.status()==0)
	    	m_tail->count_beaconsTimer.sched(b_period);
	    else
	    	printf("ERRO \n");
    
	    Refresh_group_leader();
	    double interval = BEACON_TIMEOUT;
	    m_tail->timer.resched(interval);
	    return;
	}

	// refresh seq_num because beacon address is in the table and it is received 1rst time
	if (res == 0) {
		ptr->bc_elem_seq_num = bc_seq_num;
		ptr->bc_elem_bl_associativity = bc_associativity;
		ptr->bc_elem_gl_addr = bc_gl_addr;
		this->Refresh_group_leader();
		double interval = BEACON_TIMEOUT;
		ptr->timer.resched(interval);
		return;
	}

  	// halt - the beacon was already received
  	if (res == 1)
	  return;
}

void Beacons_tab::Delete_beacon(nsaddr_t *received_from)
{ 
	Beacons_tab_node *ptr,*last_node_ptr;
	ptr = m_head;
	last_node_ptr = m_head;

	// search if address is contained in the table
	for(int i=0; i<m_elems;i++) {
		if (ptr->bc_elem_node == *received_from)
			break;
		last_node_ptr = ptr;
		ptr = ptr->next;
	}
	//
	if (ptr!=NULL) {
		last_node_ptr->next = ptr->next;  
		if (ptr == m_tail) m_tail = last_node_ptr;
		if (ptr == m_head) m_head = ptr->next;
		if (m_elems==1) {
			m_head = NULL;
			m_tail = NULL;
		}
		
		// link duration statistics
		link_counter ++;
		if ((histogram.size()) < ptr->bc_elem_associativity) //position 0 match with 1 second duration links
		{
			histogram.reserve(ptr->bc_elem_associativity);
			histogram.resize(ptr->bc_elem_associativity,0);
		}
		histogram[ptr->bc_elem_associativity-1]++;
		// end link duration statistics
    
		ptr->timer.force_cancel();
		ptr->count_beaconsTimer.force_cancel();
		delete ptr;
		m_elems--;
	}
	Refresh_group_leader();

	return;
}

int Beacons_tab::Get_num_elems()
{
	return m_elems;
}

// It gives ro_max and v_max contained in the table,
// it also returns if the node is stable or unstable
//
bool Beacons_tab::Refresh_stability(int *ro_max, int *v_max)
{
	Beacons_tab_node *ptr;
	ptr = m_head;
	*ro_max = -2;
	*v_max = -2;
	for(int i=0; i<m_elems;i++){
		if (ptr->bc_elem_associativity > *ro_max)
			*ro_max = ptr->bc_elem_associativity;
		if (ptr->bc_elem_bl_associativity > *v_max)
			*v_max = ptr->bc_elem_bl_associativity;
		ptr = ptr->next;
	}
	if (*ro_max >= BEACON_STABILITY_THRESHOLD)
		return true;
	else
        return false;
}

void Beacons_tab::insert_in_sort_list(u_int32_t ask,vector<u_int32_t> &vectr)
{
	vector<u_int32_t>::iterator itrt;

    if (vectr.size()==0)
    	vectr.push_back(ask);
    else
    	for (itrt = vectr.begin(); itrt != vectr.end(); ++itrt) {
    		if (*itrt > ask && *(itrt-1) < ask) {
    			vectr.insert(itrt,ask);
    			break;
    		}
       }
    if (ask > vectr[vectr.size()-1])
    	vectr.push_back(ask);
    if (ask < vectr[0]){
    	itrt = vectr.begin();
    	vectr.insert(itrt, ask);
    }
}

//------------------------------------------------------------------------
// Called when one beacon is inserted or removed from the table
// Its function is update all variables related with node's group leader:
// u_int16_t  bc_associativity;
// nsaddr_t   bc_gl_addr;			// -1 ==> no group leader
// u_int8_t   bc_gl_lis_elems;		// Num of elements contained in bc_gl_list - list of group leaders in the neigborhood
// char*      bc_gl_list;
//-----------------------------------------------------------------------------------------------------------------
void Beacons_tab::Refresh_group_leader()
{
	int ro_max, v_max;						// Node stability 
	Beacons_tab_node *ptr_win_node=NULL;	// Pointer to gl node entry table
	Beacons_tab_node *ptr;
	bool elected = false;
	bool auto_elected = false;
	nsaddr_t last_addr = 100000;
	//////////////////////////////////////////////////////
	// Filter - unstable nodes counter
	int unstab_thesh = 1;
	//////////////////////////////////////////////////////
	Refresh_is_BGL();
	m_is_stable = Refresh_stability(&ro_max, &v_max);

	if (m_is_stable) {
		vector<u_int32_t> BGL_lst;
		BGL_lst.clear();
		ptr = m_head;
		for(int i = 0; i<m_elems;i++) {
			if (ptr->bc_elem_gl_addr != -1)
				this->insert_in_sort_list(ptr->bc_elem_gl_addr, BGL_lst);
			ptr = ptr->next;
		}
		for(int j=0;j<(int)BGL_lst.size();j++) {
			int key = BGL_lst[j];
			ptr = m_head;
			for(int i=0; i<m_elems;i++) {
				if ((ptr->bc_elem_node == key) && (ptr->bc_elem_associativity >=BEACON_STABILITY_THRESHOLD))
				{
					elected = true;
					ptr_win_node = ptr;
					break;
				}
				ptr = ptr->next;
			}
			if (elected) break;
			ptr = m_head;
			if (address == key) {
				auto_elected = true;
				elected = true;
				ptr_win_node = ptr;
				break;
			}
		}
		ptr = m_head;
		if (!elected)
			for(int i = 0; i<m_elems;i++) {
				if ((ro_max - ptr->bc_elem_associativity - unstab_thesh <= 0) && (ptr->bc_elem_node < last_addr))
				{
					last_addr = ptr->bc_elem_node;
					ptr_win_node = ptr;
				}
				ptr = ptr->next;
			}
	}

	// Update data structures to build beacon send information
	if (ptr_win_node == NULL)//node is unstable
	{
		if (ro_max < 0) ro_max =0;
		bc_associativity = ro_max;
		bc_gl_addr = -1;
		bc_gl_lis_elems = 0;
		bc_gl_list = NULL;
    	#ifdef DEBUG_CLUSTERING
			this->Print();
    	#endif    
		return;
	}
	if (!auto_elected)		// Node elected is in the table
	{
		bc_associativity = ro_max;
		bc_gl_addr = ptr_win_node->bc_elem_node;
		bc_gl_lis_elems = 0;
		bc_gl_list = NULL;
	}
	else 					// Node elected is the table owner's
	{
		bc_associativity = ro_max;
		bc_gl_addr = this->address;
		bc_gl_lis_elems = 0;
		bc_gl_list = NULL;
	}
   	#ifdef DEBUG_CLUSTERING
    	this->Print();
   	#endif
}

void Beacons_tab::Refresh_is_BGL()
{
	Beacons_tab_node *ptr;
	ptr = m_head;

	for(int i=0; i<m_elems;i++) {
		if (this->address == ptr->bc_elem_gl_addr) {
			m_is_bgl = true;
			return;
		}
		ptr = ptr->next;
	}
	m_is_bgl = false;
}

//------------------------------------------------------------------------
// Returns associativity node, towards processing node
u_int16_t Beacons_tab::Return_assoc_of_node(nsaddr_t received_from){
	Beacons_tab_node *ptr;
	ptr = m_head;
	for(int i=0; i<m_elems;i++){
	    if (ptr->bc_elem_node == received_from)
	    	return ptr->bc_elem_associativity;
	    ptr = ptr->next;
	}
	return 0;
}

//------------------------------------------------------------------------
// Returns associativity of all nodes, towards procession node
std::map<nsaddr_t, u_int16_t> Beacons_tab::Return_assoc_of_all_nodes(){
	Beacons_tab_node *ptr;
	std::map<nsaddr_t, u_int16_t> node_assoc;
	ptr = m_head;
	for(int i=0; i<m_elems;i++){
		node_assoc.insert(make_pair(ptr->bc_elem_node, ptr->bc_elem_associativity));
	    ptr = ptr->next;
	}
	return node_assoc;
}

void Beacons_tab::Print() 
{
	Beacons_tab_node *ptr;
	ptr = m_head;

	fprintf(stdout,"-----------------------------------------------------------------\n");

	if (m_is_bgl) fprintf(stdout,"==================> is     B G L \n");
	fprintf(stdout,"Vz(%d)\t Sum(%d)\tGL(%d)\tStab(%d)\tMy_GL=%d\t\n",address,address,address,address,bc_gl_addr);
	for (int i=0; i<m_elems;i++)
	{
		fprintf(stdout,"%d\t %d\t%d\t%d\t\n", ptr->bc_elem_node,
     				  						ptr->bc_elem_associativity,
     				  						ptr->bc_elem_gl_addr,
     				  						ptr->bc_elem_bl_associativity);
		ptr = ptr->next;
	}
};

Beacons_tab_node* Beacons_tab::Get_elem_at(int pos)
{
	int i;
	Beacons_tab_node *ptr;

	if (pos > m_elems-1)
		return NULL;
	ptr = m_head;
	for (i=0; i< pos; i++) // position starts at index 0
		ptr = ptr->next;

	return ptr;
}

void  Beacons_tab::print_link_duration_stats()
{
	int i;

	printf("----------------------------------------------------------------------\n");
	printf("NODE %d\n", address);
	printf("NUMBER of Links = %d\n",link_counter);
	printf("BEGIN HISTOGRAM DESCRIPTION -->size=%d\n",histogram.size());

	for (i=0; (int)i<(int)histogram.size();i++)
		printf("%d\n",histogram[i]);
	printf("END HISTOGRAM DESCRIPTION\n");
}
